@@ -4,7 +4,7 @@
 //
 // This module parses the very complicated exif structures.
 //
-// Matthias Wandel,  Dec 1999 - January 2002
+// Matthias Wandel,  Dec 1999 - Dec 2002 
 //--------------------------------------------------------------------------
 #include <stdio.h>
 #include <stdlib.h>
@@ -34,6 +34,10 @@ static double FocalplaneUnits;
 static int ExifImageWidth;
 static int MotorolaOrder = 0;
 
+// for fixing the rotation.
+static void * OrientationPtr;
+static int    OrientationNumFormat; 
+
 typedef struct {
     unsigned short Tag;
     char * Desc;
@@ -59,6 +63,31 @@ static TagTable_t ProcessTable[] = {
     { 0,        "Unknown"}
 };
 
+// 1 - "The 0th row is at the visual top of the image,    and the 0th column is the visual left-hand side."
+// 2 - "The 0th row is at the visual top of the image,    and the 0th column is the visual right-hand side."
+// 3 - "The 0th row is at the visual bottom of the image, and the 0th column is the visual right-hand side."
+// 4 - "The 0th row is at the visual bottom of the image, and the 0th column is the visual left-hand side."
+
+// 5 - "The 0th row is the visual left-hand side of of the image,  and the 0th column is the visual top."
+// 6 - "The 0th row is the visual right-hand side of of the image, and the 0th column is the visual top."
+// 7 - "The 0th row is the visual right-hand side of of the image, and the 0th column is the visual bottom."
+// 8 - "The 0th row is the visual left-hand side of of the image,  and the 0th column is the visual bottom."
+
+// Note: The descriptions here are the same as the name of the command line
+// option to pass to jpegtran to right the image
+
+static const char * OrientTab[9] = {
+    "Undefined",
+    "Normal",           // 1
+    "flip horizontal",  // left right reversed mirror
+    "rotate 180",       // 3
+    "flip vertical",    // upside down mirror
+    "transpose",        // Flipped about top-left <--> bottom-right axis.
+    "rotate 90",        // rotate 90 cw to right it.
+    "transverse",       // flipped about top-right <--> bottom-left axis
+    "rotate 270",       // rotate 270 to right it.
+};
+
 
 //--------------------------------------------------------------------------
 // Describes format descriptor
@@ -87,6 +116,8 @@ static int BytesPerFormat[] = {0,1,1,2,4,8,1,1,2,4,8,4,8};
 #define TAG_MAKE              0x010F
 #define TAG_MODEL             0x0110
 
+#define TAG_ORIENTATION       0x0112
+
 #define TAG_EXPOSURETIME      0x829A
 #define TAG_FNUMBER           0x829D
 
@@ -112,7 +143,6 @@ static int BytesPerFormat[] = {0,1,1,2,4,8,1,1,2,4,8,4,8};
 #define TAG_METERING_MODE     0x9207
 #define TAG_EXPOSURE_PROGRAM  0x8822
 #define TAG_ISO_EQUIVALENT    0x8827
-#define TAG_COMPRESSION_LEVEL 0x9102
 
 #define TAG_THUMBNAIL_OFFSET  0x0201
 #define TAG_THUMBNAIL_LENGTH  0x0202
@@ -218,7 +248,6 @@ static void Put16u(void * Short, unsigned short PutValue)
     }
 }
 
-
 //--------------------------------------------------------------------------
 // Convert a 16 bit unsigned value from file's native byte order
 //--------------------------------------------------------------------------
@@ -260,7 +289,7 @@ static void PrintFormatNumber(void * ValuePtr, int Format, int ByteCount)
 {
     switch(Format){
         case FMT_SBYTE:
-        case FMT_BYTE:      printf("%02x ",*(uchar *)ValuePtr);             break;
+        case FMT_BYTE:      printf("%02x\n",*(uchar *)ValuePtr);            break;
         case FMT_USHORT:    printf("%d\n",Get16u(ValuePtr));                break;
         case FMT_ULONG:     
         case FMT_SLONG:     printf("%d\n",Get32s(ValuePtr));                break;
@@ -320,7 +349,8 @@ static double ConvertAnyFormat(void * ValuePtr, int Format)
 //--------------------------------------------------------------------------
 // Process one of the nested EXIF directories.
 //--------------------------------------------------------------------------
-static void ProcessExifDir(unsigned char * DirStart, unsigned char * OffsetBase, unsigned ExifLength)
+static void ProcessExifDir(unsigned char * DirStart, unsigned char * OffsetBase, 
+        unsigned ExifLength, int NestingLevel)
 {
     int de;
     int a;
@@ -328,6 +358,12 @@ static void ProcessExifDir(unsigned char * DirStart, unsigned char * OffsetBase,
     unsigned ThumbnailOffset = 0;
     unsigned ThumbnailSize = 0;
 
+    if (NestingLevel > 4){
+        ErrNonfatal("Maximum directory nesting exceeded (corrupt exif header)", 0,0);
+        return;
+    }
+
+
     NumDirEntries = Get16u(DirStart);
     #define DIR_ENTRY_ADDR(Start, Entry) (Start+2+12*(Entry))
 
@@ -341,10 +377,11 @@ static void ProcessExifDir(unsigned char * DirStart, unsigned char * OffsetBase,
             }else{
                 // Note: Files that had thumbnails trimmed with jhead 1.3 or earlier
                 // might trigger this.
-                ErrExit("Illegally sized directory");
+                ErrNonfatal("Illegally sized directory",0,0);
+                return;
             }
         }
-        if (DirEnd < LastExifRefd) LastExifRefd = DirEnd;
+        if (DirEnd > LastExifRefd) LastExifRefd = DirEnd;
     }
 
     if (ShowTags){
@@ -364,7 +401,8 @@ static void ProcessExifDir(unsigned char * DirStart, unsigned char * OffsetBase,
 
         if ((Format-1) >= NUM_FORMATS) {
             // (-1) catches illegal zero case as unsigned underflows to positive large.
-            ErrExit("Illegal format code in EXIF dir");
+            ErrNonfatal("Illegal number format %d for tag %04x", Format, Tag);
+            continue;
         }
 
         ByteCount = Components * BytesPerFormat[Format];
@@ -375,9 +413,8 @@ static void ProcessExifDir(unsigned char * DirStart, unsigned char * OffsetBase,
             // If its bigger than 4 bytes, the dir entry contains an offset.
             if (OffsetVal+ByteCount > ExifLength){
                 // Bogus pointer offset and / or bytecount value
-                printf("Offset %d bytes %d ExifLen %d\n",OffsetVal, ByteCount, ExifLength);
-
-                ErrExit("Illegal pointer offset value in EXIF");
+                ErrNonfatal("Illegal value pointer for tag %04x", Tag,0);
+                continue;
             }
             ValuePtr = OffsetBase+OffsetVal;
         }else{
@@ -389,6 +426,7 @@ static void ProcessExifDir(unsigned char * DirStart, unsigned char * OffsetBase,
             // Keep track of last byte in the exif header that was actually referenced.
             // That way, we know where the discardable thumbnail data begins.
             LastExifRefd = ValuePtr+ByteCount;
+
         }
 
         if (ShowTags){
@@ -416,12 +454,12 @@ static void ProcessExifDir(unsigned char * DirStart, unsigned char * OffsetBase,
                         int NoPrint = 0;
                         printf("\"");
                         for (a=0;a<ByteCount;a++){
-                            if (isprint((ValuePtr)[a])){
+                            if (ValuePtr[a] >= 32){
                                 putchar((ValuePtr)[a]);
                                 NoPrint = 0;
                             }else{
                                 // Avoiding indicating too many unprintable characters of proprietary
-                                // bits of binary information this program may know how to parse.
+                                // bits of binary information this program may not know how to parse.
                                 if (!NoPrint){
                                     putchar('?');
                                     NoPrint = 1;
@@ -526,8 +564,20 @@ static void ProcessExifDir(unsigned char * DirStart, unsigned char * OffsetBase,
                 break;
 
             case TAG_FLASH:
-                if (ConvertAnyFormat(ValuePtr, Format)){
-                    ImageInfo.FlashUsed = 1;
+                if ((int)ConvertAnyFormat(ValuePtr, Format) & 7){
+                    ImageInfo.FlashUsed = TRUE;
+                }else{
+                    ImageInfo.FlashUsed = FALSE;
+                }
+                break;
+
+            case TAG_ORIENTATION:
+                ImageInfo.Orientation = (int)ConvertAnyFormat(ValuePtr, Format);
+                OrientationPtr = ValuePtr;
+                OrientationNumFormat = Format;
+                if (ImageInfo.Orientation < 1 || ImageInfo.Orientation > 8){
+                    ErrNonfatal("Undefined rotation value %d", ImageInfo.Orientation, 0);
+                    ImageInfo.Orientation = 0;
                 }
                 break;
 
@@ -582,10 +632,6 @@ static void ProcessExifDir(unsigned char * DirStart, unsigned char * OffsetBase,
                 if ( ImageInfo.ISOequivalent < 50 ) ImageInfo.ISOequivalent *= 200;
                 break;
 
-            case TAG_COMPRESSION_LEVEL:
-                ImageInfo.CompressionLevel = (int)ConvertAnyFormat(ValuePtr, Format);
-                break;
-
             case TAG_THUMBNAIL_OFFSET:
                 ThumbnailOffset = (unsigned)ConvertAnyFormat(ValuePtr, Format);
                 DirWithThumbnailPtrs = DirStart;
@@ -601,9 +647,9 @@ static void ProcessExifDir(unsigned char * DirStart, unsigned char * OffsetBase,
                     unsigned char * SubdirStart;
                     SubdirStart = OffsetBase + Get32u(ValuePtr);
                     if (SubdirStart < OffsetBase || SubdirStart > OffsetBase+ExifLength){
-                        ErrExit("Illegal subdirectory link 1");
+                        ErrNonfatal("Illegal exif or interop ofset directory link",0,0);
                     }else{
-                        ProcessExifDir(SubdirStart, OffsetBase, ExifLength);
+                        ProcessExifDir(SubdirStart, OffsetBase, ExifLength, NestingLevel+1);
                     }
                     continue;
                 }
@@ -629,11 +675,11 @@ static void ProcessExifDir(unsigned char * DirStart, unsigned char * OffsetBase,
                         // I'll just let it pass silently
                         if (ShowTags) printf("Thumbnail removed with Jhead 1.3 or earlier\n");
                     }else{
-                        ErrExit("Illegal subdirectory link 2");
+                        ErrNonfatal("Illegal subdirectory link",0,0);
                     }
                 }else{
                     if (SubdirStart <= OffsetBase+ExifLength){
-                        ProcessExifDir(SubdirStart, OffsetBase, ExifLength);
+                        ProcessExifDir(SubdirStart, OffsetBase, ExifLength, NestingLevel+1);
                     }
                 }
             }
@@ -662,11 +708,13 @@ static void ProcessExifDir(unsigned char * DirStart, unsigned char * OffsetBase,
 //--------------------------------------------------------------------------
 void process_EXIF (unsigned char * ExifSection, unsigned int length)
 {
-    ImageInfo.FlashUsed = 0; // If it s from a digicam, and it used flash, it says so.
+    int FirstOffset;
 
     FocalplaneXRes = 0;
     FocalplaneUnits = 0;
     ExifImageWidth = 0;
+    OrientationPtr = NULL;
+
 
     if (ShowTags){
         printf("Exif header %d bytes long\n",length);
@@ -675,7 +723,8 @@ void process_EXIF (unsigned char * ExifSection, unsigned int length)
     {   // Check the EXIF header component
         static uchar ExifHeader[] = "Exif\0\0";
         if (memcmp(ExifSection+2, ExifHeader,6)){
-            ErrExit("Incorrect Exif header");
+            ErrNonfatal("Incorrect Exif header",0,0);
+            return;
         }
     }
 
@@ -687,21 +736,29 @@ void process_EXIF (unsigned char * ExifSection, unsigned int length)
             if (ShowTags) printf("Exif section in Motorola order\n");
             MotorolaOrder = 1;
         }else{
-            ErrExit("Invalid Exif alignment marker.");
+            ErrNonfatal("Invalid Exif alignment marker.",0,0);
+            return;
         }
     }
 
-    // Check the next two values for correctness.
-    if (Get16u(ExifSection+10) != 0x2a
-      || Get32u(ExifSection+12) != 0x08){
-        ErrExit("Invalid Exif start (1)");
+    // Check the next value for correctness.
+    if (Get16u(ExifSection+10) != 0x2a){
+        ErrNonfatal("Invalid Exif start (1)",0,0);
+        return;
+    }
+
+    FirstOffset = Get32u(ExifSection+12);
+    if (FirstOffset < 8 || FirstOffset > 16){
+        // I used to ensure this was set to 8 (website I used indicated its 8)
+        // but PENTAX Optio 230 has it set differently, and uses it as offset. (Sept 11 2002)
+        ErrNonfatal("Suspicious offset of first IFD value",0,0);
     }
 
     LastExifRefd = ExifSection;
     DirWithThumbnailPtrs = NULL;
 
     // First directory starts 16 bytes in.  All offset are relative to 8 bytes in.
-    ProcessExifDir(ExifSection+16, ExifSection+8, length-6);
+    ProcessExifDir(ExifSection+8+FirstOffset, ExifSection+8, length-6, 0);
 
     // Compute the CCD width, in milimeters.
     if (FocalplaneXRes != 0){
@@ -713,6 +770,42 @@ void process_EXIF (unsigned char * ExifSection, unsigned int length)
     }
 }
 
+//--------------------------------------------------------------------------
+// Cler the rotation tag in the exif header to 1.
+// Note: The thumbnail is NOT rotated.  That would be really hard, especially
+// stuffing it back into the exif header, because its size might change.
+//--------------------------------------------------------------------------
+const char * ClearOrientation(void)
+{
+    if (OrientationPtr == NULL) return NULL;
+    
+    switch(OrientationNumFormat){
+        case FMT_SBYTE:
+        case FMT_BYTE:      
+            *(uchar *)OrientationPtr = 1;
+            break;
+
+        case FMT_USHORT:    
+            Put16u(OrientationPtr, 1);                
+            break;
+
+        case FMT_ULONG:     
+        case FMT_SLONG:     
+            memset(OrientationPtr, 0, 4);
+            // Can't be bothered to write  generic Put32 if I only use it once.
+            if (MotorolaOrder){
+                ((uchar *)OrientationPtr)[3] = 1;
+            }else{
+                ((uchar *)OrientationPtr)[0] = 1;
+            }
+            break;
+
+        default:
+            return NULL;
+    }
+
+    return OrientTab[ImageInfo.Orientation];
+}
 
 
 //--------------------------------------------------------------------------
@@ -811,6 +904,12 @@ void ShowImageInfo(void)
         printf("Date/Time    : %s\n",ImageInfo.DateTime);
     }
     printf("Resolution   : %d x %d\n",ImageInfo.Width, ImageInfo.Height);
+
+    if (ImageInfo.Orientation > 1){
+        // Only print orientation if one was supplied, and if its not 1 (normal orientation)
+        printf("Orientation  : %s\n", OrientTab[ImageInfo.Orientation]);
+    }
+
     if (ImageInfo.IsColor == 0){
         printf("Color/bw     : Black and white\n");
     }
@@ -821,17 +920,21 @@ void ShowImageInfo(void)
         printf("Focal length : %4.1fmm",(double)ImageInfo.FocalLength);
         if (ImageInfo.CCDWidth){
             printf("  (35mm equivalent: %dmm)",
-                        (int)(ImageInfo.FocalLength/ImageInfo.CCDWidth*35 + 0.5));
+                        (int)(ImageInfo.FocalLength/ImageInfo.CCDWidth*36 + 0.5));
         }
         printf("\n");
     }
 
     if (ImageInfo.CCDWidth){
-        printf("CCD Width    : %4.2fmm\n",(double)ImageInfo.CCDWidth);
+        printf("CCD width    : %4.2fmm\n",(double)ImageInfo.CCDWidth);
     }
 
     if (ImageInfo.ExposureTime){
-        printf("Exposure time:%6.3f s ",(double)ImageInfo.ExposureTime);
+        if (ImageInfo.ExposureTime < 0.010){
+            printf("Exposure time: %6.4f s ",(double)ImageInfo.ExposureTime);
+        }else{
+            printf("Exposure time: %5.3f s ",(double)ImageInfo.ExposureTime);
+        }
         if (ImageInfo.ExposureTime <= 0.5){
             printf(" (1/%d)",(int)(0.5 + 1/ImageInfo.ExposureTime));
         }
@@ -842,9 +945,9 @@ void ShowImageInfo(void)
     }
     if (ImageInfo.Distance){
         if (ImageInfo.Distance < 0){
-            printf("Focus Dist.  : Infinite\n");
+            printf("Focus dist.  : Infinite\n");
         }else{
-            printf("Focus Dist.  : %4.2fm\n",(double)ImageInfo.Distance);
+            printf("Focus dist.  : %4.2fm\n",(double)ImageInfo.Distance);
         }
     }
 
@@ -897,21 +1000,6 @@ void ShowImageInfo(void)
             break;
         }
     }
-    if (ImageInfo.CompressionLevel){ // 05-jan-2001 vcs
-        switch(ImageInfo.CompressionLevel) {
-        case 1:
-            printf("JPG Quality  : basic\n");
-            break;
-        case 2:
-            printf("JPG Quality  : normal\n");
-            break;
-        case 4:
-            printf("JPG Quality  : fine\n");
-            break;
-       }
-    }
-
-         
 
     for (a=0;;a++){
         if (ProcessTable[a].Tag == ImageInfo.Process || ProcessTable[a].Tag == 0){
@@ -940,8 +1028,8 @@ void ShowImageInfo(void)
                 putchar(c);
             }
         }
+        printf("\n");
     }
-
     printf("\n");
 }
 
